print("import: 'libpysal'")
import libpysal

print("import: 'libpysal.cg'")
import libpysal.cg

print("import: 'libpysal.cg.ops'")
import libpysal.cg.ops

print("import: 'libpysal.examples'")
import libpysal.examples

print("import: 'libpysal.io'")
import libpysal.io

print("import: 'libpysal.io.geotable'")
import libpysal.io.geotable

print("import: 'libpysal.io.iohandlers'")
import libpysal.io.iohandlers

print("import: 'libpysal.io.tests'")
import libpysal.io.tests

print("import: 'libpysal.io.util'")
import libpysal.io.util

print("import: 'libpysal.weights'")
import libpysal.weights

